# controller.py : skeleton Webots API to connect the controller objects to the physical hardware.

import time, threading, queue

################################################################
class Robot(object):
    """This object emulates the portions of the Webots API common to all physical robots."""
    def __init__(self):

        print("Entering Robot.__init__ and initializing superclasses")
        super(Robot, self).__init__()

        # include a receiver/emitter pair by default
        self.emitter = Emitter()
        self.receiver = Receiver()

        # lock to synchronize access for this object and its children
        self._lock = threading.Lock()

    def step(self, interval):
        #print(f"Robot step function sleeping for {interval} msec on {self}.")

        self._lock.release()
        time.sleep(0.001 * interval)
        self._lock.acquire()

        #print(f"Robot step function resuming.")
        return 0

    # only support a single receiver and emitter
    def getReceiver(self, name):
        return self.receiver

    def getEmitter(self, name):
        return self.emitter

################################################################
class StudentRobot(object):
    """This object emulates the portions of the Webots API related to the physical student robot."""
    def __init__(self):
        print("Entering StudentRobot.__init__ and initializing superclasses")
        super(StudentRobot, self).__init__()

        # Create objects representing the physical hardware.
        self._sensors = { name : RealSensor() for name in ['sensor0', 'sensor1'] }
        self._motors  = { name : RealMotor()  for name in ['motor0', 'motor1'] }

    def getPositionSensor(self, name):
        return self._sensors[name]

    def getMotor(self, name):
        return self._motors[name]

    def process_sensor_string(self, line):
        # this is typically called from a serial port thread with data from the Arduino
        try:
            values = [int(s) for s in line.split()]
            if len(values) >= 2:
                with self._lock:
                    self._sensors['sensor0'].value = values[0] #this is unchanged since python doesn't do anything with speed
                    self._sensors['sensor1'].value = values[1]
        except ValueError:
            pass

    def get_output_string(self):
        # this is typically called from a serial port thread to generate output for the Arduino
        with self._lock:
            p0 = int(self._motors['motor0'].target) # being sent to arduino
            v0 = int(self._motors['motor0'].rate)
            p1 = int(self._motors['motor1'].target)
            v1 = int(self._motors['motor1'].rate)
            s = f"{p0} {v0} {p1} {v1}\n"
        return s

################################################################
class TutorRobot(object):
    """This object emulates the portions of the Webots API related to the physical tutor robot."""
    def __init__(self):
        print("Entering TutorRobot.__init__ and initializing superclasses")
        super(TutorRobot, self).__init__()

        # Create objects representing the physical hardware.
        self._sensors = { name : RealSensor() for name in ['sensor0', 'sensor1', 'sensor2'] }
        self._motors  = { name : RealMotor()  for name in ['motor0', 'motor1', 'motor2'] }

    def getPositionSensor(self, name):
        return self._sensors[name]

    def getMotor(self, name):
        return self._motors[name]

    def process_sensor_string(self, line):
        # this is typically called from a serial port thread with data from the Arduino
        try:
            values = [int(s) for s in line.split()]
            if len(values) >= 3: #IF THINGS GO WRONG MAYBE THIS SHOULD BE 4 NOT 3
                with self._lock:
                    self._sensors['sensor0'].value = values[0] #this is unchanged since python doesn't do anything with speed
                    self._sensors['sensor1'].value = values[1]
                    self._sensors['sensor2'].value = values[2]
        except ValueError:
            pass

    def get_output_string(self):
        # this is typically called from a serial port thread to generate output for the Arduino
        with self._lock:
            p0 = int(self._motors['motor0'].target) # being sent to arduino
            v0 = int(self._motors['motor0'].rate)
            p1 = int(self._motors['motor1'].target)
            v1 = int(self._motors['motor1'].rate)
            p2 = int(self._motors['motor2'].target)
            v2 = int(self._motors['motor2'].rate)
            s = f"{p0} {v0} {p1} {v1} {p2} {v2}\n"
        return s

################################################################
class RealSensor(object):
    def __init__(self):
        self.value = 0

    def enable(self, rate):
        pass

    def getValue(self):
        return self.value

################################################################
class RealMotor(object):
    def __init__(self):
        self.rate   = 0
        self.target = 0

    def setVelocity(self, rate):
        self.rate = rate

    def setPosition(self, target):
        self.target = target

################################################################
class Receiver(object):
    def __init__(self):

        # create a thread-safe queue for received messages
        self.queue = queue.Queue()
        pass

    def enable(self, rate):
        pass

    def getQueueLength(self):
        # return the number of available packets
        return self.queue.qsize()

    def getData(self):
        # return the next packet data as a bytestring
        # this also removes it from the queue, which doesn't match the Webots behavior, but
        # will work as long as it isn't called twice
        data = self.queue.get()
        #print(f"receiver returning {data}")
        return data

    def nextPacket(self):
        # remove the oldest packet from the queue
        # because it is removed by getData, this is a no-op
        pass

class Emitter(object):
    def __init__(self):
        self.receivers = []

    def add_receiver(self, receiver):
        self.receivers.append(receiver)

    def send(self, data):
        # send a bytestring to the subscribed receivers
        #print(f"emitter broadcasting {data}")
        for r in self.receivers:
            r.queue.put(data)

################################################################
