# performance.py : run the piano tutor show

# This provides a skeleton Webots API to connect the controller objects to the
# physical hardware.  Unlike Webots, it uses threads to run the controllers as
# normal Python objects within a single process.

# Standard Python modules.
import os, sys, time, threading

# pySerial
import serial

# Import the skeleton Webots API from the current folder.
import controller

# Import the controllers.  This program assumes a directory structure where it
# exists within the Webots project tree.  The following configures the load path
# both to load the controller modules and to ensure they can load the skeleton
# API.
parent_path = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
controller_path = os.path.join(parent_path, "controllers")
print(f"Controller modules will load from {controller_path}")
sys.path.insert(0, controller_path)

import modified_student.modified_student as student
import modified_tutor.modified_tutor     as tutor

################################################################
# Define wrapper classes which inherit from both a Webots robot controller and a
# specialized physical interface class.  The controller object inherits
# controller.Robot with common API code, but some methods are implemented
# instead by the interface class.

class RealStudent(student.Student, controller.StudentRobot, threading.Thread):
    def __init__(self):
        print("Entering RealStudent.__init__ and initializing superclasses")
        super(RealStudent, self).__init__()

    def run(self):
        print("Starting student.")
        self._lock.acquire()
        super(RealStudent, self).run()

class RealTutor(tutor.Tutor, controller.TutorRobot, threading.Thread):
    def __init__(self):
        print("Entering RealTutor.__init__ and initializing superclasses")
        super(RealTutor, self).__init__()

    def run(self):
        print("Starting tutor.")
        self._lock.acquire()
        super(RealTutor, self).run()

################################################################
class Arduino(threading.Thread):
    """Connect an Arduino serial port to a controller object.  The input and output
    is assumed to be a text line of space-delimited integers. The specific data
    format of both input is determined by the controller class and the Arduino
    sketch.  This process runs as a thread so it can easily block on the input.
    """
    def __init__(self, portname, controller):
        print("Entering Arduino.__init__ and initializing superclasses")
        super(Arduino, self).__init__()

        self.controller = controller
        self.portname = portname
        self.port = serial.Serial(portname, baudrate=9600, timeout=1.0)

    def run(self):
        # wait briefly for the Arduino to complete waking up
        time.sleep(0.2)

        print(f"Entering Arduino event loop for {self.portname}.")
        while(True):
            #print("running")
            line = self.controller.get_output_string()
            self.port.write(line.encode())
            print(f"Sending to Arduino: {line}")

            input = self.port.readline().decode(encoding='ascii',errors='ignore')
            #print(f"Received from Arduino: {input}")
            self.controller.process_sensor_string(input)

################################################################
student = RealStudent()
tutor = RealTutor()

# cross-connect the emitters and receivers
student.emitter.add_receiver(tutor.receiver)
tutor.emitter.add_receiver(student.receiver)

student_hw = Arduino("COM6", student)
tutor_hw = Arduino("COM4", tutor)


student.start()
tutor.start()

student_hw.start()
tutor_hw.start()

student_hw.join()
tutor_hw.join()

